@@ -1,4 +1,7 @@
-import os
+import os, sys, stat
+from pathlib import Path
+import shutil
+import glob
 import click
 import configparser
 import datetime
@@ -19,7 +22,7 @@ docker_compose_cmd = 'docker-compose -f d2s-cwl-workflows/docker-compose.yaml '
 def get_services_list(ctx, args, incomplete):
     # TODO: automate by parsing the docker-compose.yaml
     return filter(lambda x: x.startswith(incomplete), [ 'demo',
-    'graphdb', 'virtuoso', 'tmp-virtuoso', 'blazegraph', 'allegrograph', 'anzograph',
+    'graphdb', 'virtuoso', 'tmp-virtuoso', 'blazegraph', 'allegrograph', 'anzograph', 'fuseki',
     'into-the-graph', 'ldf-server', 'comunica', 'notebook',
     'api', 'drill', 'postgres', 'proxy', 'filebrowser', 'rmlstreamer', 'rmltask',
     'neo4j' ])
@@ -29,17 +32,28 @@ def get_workflows_list(ctx, args, incomplete):
     return filter(lambda x: x.startswith(incomplete), os.listdir("./d2s-cwl-workflows/workflows"))
 def get_workflow_history(ctx, args, incomplete):
     # Sorted from latest to oldest
-    files = list(filter(lambda x: x.startswith(incomplete), os.listdir("./workspace/workflow-history")))
-    return sorted(files, key=lambda x: os.path.getmtime("./workspace/workflow-history/" + x), reverse=True)
+    files = list(filter(lambda x: x.startswith(incomplete), os.listdir("./workspace/logs")))
+    return sorted(files, key=lambda x: os.path.getmtime("./workspace/logs/" + x), reverse=True)
 def get_running_workflows(ctx, args, incomplete):
     # Only show workflow logs that have been modified in the last 2 minutes
     # TODO: Will not work for workflow with SPARQL queries longer than 2 minutes
-    files = filter(lambda x: x.startswith(incomplete), os.listdir("./workspace/workflow-history"))
-    return filter(lambda x: datetime.datetime.fromtimestamp(os.path.getmtime("./workspace/workflow-history/" + x)) > (datetime.datetime.now() - datetime.timedelta(minutes=2)), files)
+    files = filter(lambda x: x.startswith(incomplete), os.listdir("./workspace/logs"))
+    return filter(lambda x: datetime.datetime.fromtimestamp(os.path.getmtime("./workspace/logs/" + x)) > (datetime.datetime.now() - datetime.timedelta(minutes=2)), files)
 def get_running_processes(ctx, args, incomplete):
     # Show running processes to be stopped
     return [os.system("ps ax | grep -v time | grep '[c]wl-runner' | awk '{print $1}'")]
 
+def chmod777(path):
+    for dirpath, dirnames, filenames in os.walk(path):
+        os.chmod(dirpath, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
+        # os.chmod(dirpath, 0o777)
+        # os.chmod('workspace/import', stat.S_IRWXO)
+        # shutil.chown(dirpath, owner)
+        for filename in filenames:
+            os.chmod(os.path.join(dirpath, filename), stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
+            # os.chmod(os.path.join(dirpath, filename), 0o777)
+            # shutil.chown(os.path.join(dirpath, filename), owner)
+
 @cli.command()
 @click.argument('projectname', nargs=1)
 @click.pass_context
@@ -70,21 +84,33 @@ def init(ctx, projectname):
         os.system('git submodule add --recursive https://github.com/MaastrichtU-IDS/d2s-cwl-workflows.git')
 
     # Create workspace directories
-    os.system('mkdir -p workspace/output/tmp-outdir')
-    os.system('mkdir -p workspace/graphdb-import')
-    os.system('chmod -R 777 workspace/graphdb-import')
-    os.system('mkdir -p workspace/workflow-history')
-    os.system('mkdir -p workspace/download/releases/1')
-    os.system('cp ~/RMLStreamer.jar workspace/RMLStreamer.jar')
-    if not os.path.exists('/home/vemonet/RMLStreamer.jar'):
+    os.makedirs('workspace/output/tmp-outdir', exist_ok=True)
+    os.makedirs('workspace/output/tmp-outdir', exist_ok=True)
+    os.makedirs('workspace/import', exist_ok=True)
+    os.makedirs('workspace/logs', exist_ok=True)
+    os.makedirs('workspace/dumps/rdf/releases/1', exist_ok=True)
+    os.makedirs('workspace/dumps/hdt', exist_ok=True)
+    # Linux only, use os.chmod
+    # os.chmod(path, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
+    # https://stackoverflow.com/questions/16249440/changing-file-permission-in-python
+    chmod777('workspace/import')
+    # Read, write, and execute by others
+    # os.system('chmod -R 777 workspace/import')
+    # Get RMLStreamer from home dir to qvoid download each time
+    user_home_dir = str(Path.home())
+    if os.path.exists(user_home_dir + '/RMLStreamer.jar'):
+        shutil.copyfile(user_home_dir + '/RMLStreamer.jar', 'workspace/RMLStreamer.jar')
+    else:
         click.echo(click.style('[d2s]', bold=True) + ' Downloading RMLStreamer.jar... [80M]')
         urllib.request.urlretrieve ("https://github.com/vemonet/RMLStreamer/raw/fix-mainclass/target/RMLStreamer-1.2.2.jar", "workspace/RMLStreamer.jar")
     # os.system('wget -a workspace/RMLStreamer.jar https://github.com/vemonet/RMLStreamer/raw/fix-mainclass/target/RMLStreamer-1.2.2.jar')
-    os.system('chmod +x workspace/RMLStreamer.jar')
+    chmod777('workspace/RMLStreamer.jar')
 
     # Copy load.sh in workspace for Virtuoso bulk load
-    os.system('mkdir -p workspace/virtuoso && cp d2s-cwl-workflows/support/virtuoso/load.sh workspace/virtuoso')
-    os.system('mkdir -p workspace/tmp-virtuoso && cp d2s-cwl-workflows/support/virtuoso/load.sh workspace/tmp-virtuoso')
+    os.makedirs('workspace/virtuoso', exist_ok=True)
+    os.makedirs('workspace/tmp-virtuoso', exist_ok=True)
+    shutil.copyfile('d2s-cwl-workflows/support/virtuoso/load.sh', 'workspace/virtuoso/load.sh')
+    shutil.copyfile('d2s-cwl-workflows/support/virtuoso/load.sh', 'workspace/tmp-virtuoso/load.sh')
     # TODO: improve this to include it in Docker deployment
 
     click.echo()
@@ -95,7 +121,12 @@ def init(ctx, projectname):
         + ' and provide your email to receive the URL to download ' + click.style('GraphDB version 9.1.1 standalone zip', bold=True))
     
     graphdb_path = click.prompt(click.style('[?]', bold=True) + ' Enter the path to the GraphDB distribution 9.1.1 zip file used to build the Docker image. Default', default='~/graphdb-free-9.1.1-dist.zip')
-    os.system('cp ' + graphdb_path + ' ./d2s-cwl-workflows/support/graphdb')
+    # os.system('cp ' + graphdb_path + ' ./d2s-cwl-workflows/support/graphdb')
+    if os.path.exists(graphdb_path):
+        shutil.copyfile(graphdb_path, './d2s-cwl-workflows/support/graphdb')
+    else:
+        click.echo(click.style('[d2s]', bold=True) + ' GraphDB installation file not found. Copy the zip file in d2s-cwl-workflows/support/graphdb after download.')
+
     
     with open('.d2sconfig', 'w') as configfile:
         config.write(configfile)
@@ -113,13 +144,27 @@ def init(ctx, projectname):
 
 
 @cli.command()
-def update():
+@click.argument('services', nargs=-1, autocompletion=get_services_list)
+@click.option(
+    '--permissions/--images', default=False, 
+    help='Update files permissions (Docker images by default)')
+def update(services, permissions):
     """Update Docker images"""
-    os.system(docker_compose_cmd + 'pull')
-    os.system(docker_compose_cmd + 'build graphdb')
-    click.echo(click.style('[d2s]', bold=True) + ' All images pulled and built.')
-    click.echo(click.style('[d2s]', bold=True) + ' You can now start services (e.g. demo services):')
-    click.secho('d2s start demo', bold=True)
+    if permissions:
+        click.echo(click.style('[d2s]', bold=True) + ' Password will be asked to updates following folder permissions in workspace: input, output, import, tmp-virtuoso')
+        os.system('sudo chmod -R 777 workspace/input')
+        os.system('sudo chmod -R 777 workspace/output')
+        os.system('sudo chmod -R 777 workspace/import')
+        os.system('sudo chmod -R 777 workspace/dumps')
+        os.system('sudo chmod -R 777 workspace/tmp-virtuoso')
+    else:
+        services_string = " ".join(services)
+        os.system(docker_compose_cmd + 'pull ' + services_string)
+        if not services or "graphdb" in services:
+            os.system(docker_compose_cmd + 'build graphdb')
+        click.echo(click.style('[d2s]', bold=True) + ' All images pulled and built.')
+        click.echo(click.style('[d2s]', bold=True) + ' You can now start services (e.g. demo services):')
+        click.secho('d2s start demo', bold=True)
 
 
 @cli.command()
@@ -145,9 +190,9 @@ def config():
     help='Use custom deployment config')
 def start(services, deploy):
     """Start services"""
-    if services[0] == "demo":
+    if not services or services[0] == "demo":
         deploy = "demo"
-        services_string = "graphdb tmp-virtuoso drill api into-the-graph"
+        services_string = "graphdb tmp-virtuoso drill"
         click.echo(click.style('[d2s] ', bold=True) + ' Starting the services for the demo: ' + services_string)
     else:
         services_string = " ".join(services)
@@ -167,6 +212,7 @@ def start(services, deploy):
             click.echo(click.style('[d2s] ', bold=True) 
                 + 'Creating the repository, it should take about 20s.')
             time.sleep(10)
+            # TODO: use urllib
             os.system('curl -X POST http://localhost:7200/rest/repositories -F "config=@d2s-cwl-workflows/support/graphdb-repo-config.ttl" -H "Content-Type: multipart/form-data"')
     click.echo()
     click.echo(click.style('[d2s]', bold=True) 
@@ -183,7 +229,6 @@ def stop(services, all):
     """Stop services (--all to stop all services)"""
     if all:
         os.system(docker_compose_cmd + 'down')
-        os.system('sudo rm -rf workspace/virtuoso')
         click.echo(click.style('[d2s] ', bold=True) + 'All services stopped.')
     else:
         services_string = " ".join(services)
@@ -218,14 +263,14 @@ def process_stop(process):
 @click.argument('workflow', autocompletion=get_running_workflows)
 def watch(workflow):
     """Watch running workflow"""
-    os.system('watch tail -n 30 workspace/workflow-history/' + workflow)
+    os.system('watch tail -n 30 workspace/logs/' + workflow)
 
 
 @cli.command()
 @click.argument('workflow', autocompletion=get_workflow_history)
 def log(workflow):
     """Display logs of a workflow"""
-    os.system('less +G workspace/workflow-history/' + workflow)
+    os.system('less +G workspace/logs/' + workflow)
 
 @cli.command()
 @click.argument('datasets', nargs=-1, autocompletion=get_datasets_list)
@@ -248,13 +293,16 @@ def download(datasets):
 
 @cli.command()
 @click.argument('dataset', autocompletion=get_datasets_list)
+@click.option(
+    '--mapper/--streamer', default=False, 
+    help='Run RML Streamer or Mapper')
 @click.option(
     '--detached/--watch', default=True, 
     help='Run in detached mode or watch workflow')
-# @click.option(
-#     '-p', '--parallelism', default='8', # Not working properly.
-#     help='Run in parallel, depends on Task Slots availables')
-def rml(dataset, detached):
+@click.option(
+    '-p', '--parallelism', default='8',
+    help='Run in parallel, depends on Task Slots availables')
+def rml(dataset, detached, mapper, parallelism):
     """Run RML Streamer"""
     if (detached):
         detached_arg = "-d"
@@ -269,18 +317,27 @@ def rml(dataset, detached):
         click.echo(click.style('[d2s]', bold=True) + ' Execute mappings from ' 
             + click.style(mapping_filepath, bold=True))
     
-        rmlstreamer_cmd = 'docker exec ' + detached_arg + ' d2s-cwl-workflows_rmlstreamer_1 /opt/flink/bin/flink run -c io.rml.framework.Main /mnt/workspace/RMLStreamer.jar --path /mnt/datasets/' + dataset + '/mapping/' + mapping_filename + ' --outputPath /mnt/workspace/graphdb-import/' + mapping_filename.replace('.', '_') + '-' + dataset + '.nt --job-name "[d2s] RMLStreamer ' + mapping_filename + ' - ' + dataset + '"'
+        # TODO: store logs in a file and get the run time
+        if mapper:
+            output_filename = 'rmlmapper-' + mapping_filename.replace('.', '_') + '-' + dataset + '.nt'
+            rmlstreamer_cmd = 'docker run ' + detached_arg + ' -v $(pwd)/workspace:/mnt/workspace -v $(pwd)/datasets:/mnt/datasets umids/rmlmapper:4.7.0 -m /mnt/datasets/' + dataset + '/mapping/' + mapping_filename + ' -o /mnt/workspace/import/' + output_filename
+        else:
+            # Use RMLStreamer
+            output_filename = 'rmlstreamer-' + mapping_filename.replace('.', '_') + '-' + dataset + '.nt'
+            rmlstreamer_cmd = 'docker exec ' + detached_arg + ' d2s-cwl-workflows_rmlstreamer_1 /opt/flink/bin/flink run -c io.rml.framework.Main /mnt/workspace/RMLStreamer.jar --path /mnt/datasets/' + dataset + '/mapping/' + mapping_filename + ' --outputPath /mnt/workspace/import/' + output_filename + ' --job-name "[d2s] RMLStreamer ' + mapping_filename + ' - ' + dataset + '"'
+            print(rmlstreamer_cmd)
+            ## Try parallelism:
+            # rmlstreamer_cmd = 'docker exec ' + detached_arg + ' d2s-cwl-workflows_rmlstreamer_1 /opt/flink/bin/flink run -p ' + parallelism + ' -c io.rml.framework.Main /mnt/workspace/RMLStreamer.jar --path /mnt/datasets/' + dataset + '/mapping/' + mapping_filename + ' --outputPath /mnt/workspace/import/' + output_filename + ' --job-name "[d2s] RMLStreamer ' + mapping_filename + ' - ' + dataset + '" --parallelism ' + parallelism + ' --enable-local-parallel'
     
         os.system(rmlstreamer_cmd)
-        click.echo(click.style('[d2s]', bold=True) + ' Output file in ' )
-        click.secho('workspace/graphdb-import/'+ mapping_filename.replace('.', '_') + '-' + dataset + '.nt', bold=True)
-    
-    click.echo(click.style('[d2s]', bold=True) + ' Check the jobs running at ' 
-            + click.style('http://localhost:8078/#/job/running', bold=True))
+        click.echo(click.style('[d2s]', bold=True) + ' Output file in ')
+        click.secho('workspace/import/' + output_filename, bold=True)
+        click.echo(click.style('[d2s]', bold=True) + ' Check the jobs running at ' 
+                + click.style('http://localhost:8078/#/job/running', bold=True))
     # Try parallelism:
-    # rmlstreamer_cmd = 'docker exec -d d2s-cwl-workflows_rmlstreamer_1 /opt/flink/bin/flink run -p 4 /mnt/workspace/RMLStreamer.jar --path /mnt/datasets/' + dataset + '/mapping/rml-mappings.ttl --outputPath /mnt/workspace/graphdb-import/rml-output-' + dataset + '.nt --job-name "[d2s] RMLStreamer ' + dataset + '" --enable-local-parallel'
+    # rmlstreamer_cmd = 'docker exec -d d2s-cwl-workflows_rmlstreamer_1 /opt/flink/bin/flink run -p 4 /mnt/workspace/RMLStreamer.jar --path /mnt/datasets/' + dataset + '/mapping/rml-mappings.ttl --outputPath /mnt/workspace/import/rml-output-' + dataset + '.nt --job-name "[d2s] RMLStreamer ' + dataset + '" --enable-local-parallel'
     # Try to use docker-compose, but exec dont resolve from the -f file
-    # rmlstreamer_cmd = docker_compose_cmd + 'exec -d rmlstreamer /opt/flink/bin/flink run /mnt/workspace/RMLStreamer.jar --path /mnt/datasets/' + dataset + '/mapping/rml-mappings.ttl --outputPath /mnt/workspace/graphdb-import/rml-output-' + dataset + '.nt --job-name "[d2s] RMLStreamer ' + dataset + '"'
+    # rmlstreamer_cmd = docker_compose_cmd + 'exec -d rmlstreamer /opt/flink/bin/flink run /mnt/workspace/RMLStreamer.jar --path /mnt/datasets/' + dataset + '/mapping/rml-mappings.ttl --outputPath /mnt/workspace/import/rml-output-' + dataset + '.nt --job-name "[d2s] RMLStreamer ' + dataset + '"'
     # print(rmlstreamer_cmd)
 
 @cli.command()
@@ -311,12 +368,15 @@ def run(workflow, dataset, get_mappings, detached):
     # Virtuoso unable to handle successive bulk load + permission issues + load.sh in the virtuoso containers
     # I don't know how, they managed to not put it in the container... They had one job...
 
-    # Delete previous output (not archived)
-    os.system('rm -r workspace/output/*')
-    os.system('rm -r workspace/tmp-virtuoso/*.nq')
+    # Delete previous output (not archived). See article: https://thispointer.com/python-how-to-delete-a-directory-recursively-using-shutil-rmtree/
+    shutil.rmtree('workspace/output', ignore_errors=True, onerror=None)
+    for file in glob.glob("workspace/tmp-virtuoso/*.nq"):
+        os.remove(file)
     # Make sure the load.sh script is in the tmp Virtuoso folder
-    os.system('sudo chmod -R 777 workspace/tmp-virtuoso')
     os.system('mkdir -p workspace/tmp-virtuoso && cp d2s-cwl-workflows/support/virtuoso/load.sh workspace/tmp-virtuoso')
+    os.makedirs('workspace/tmp-virtuoso', exist_ok=True)
+    shutil.copyfile('d2s-cwl-workflows/support/virtuoso/load.sh', 'workspace/virtuoso/load.sh')
+    shutil.copyfile('d2s-cwl-workflows/support/virtuoso/load.sh', 'workspace/tmp-virtuoso/load.sh')
     
     if (detached):
         cwl_command = 'nohup time '
@@ -325,7 +385,7 @@ def run(workflow, dataset, get_mappings, detached):
     cwl_command = cwl_command +'cwl-runner --custom-net d2s-cwl-workflows_network --outdir {0}/output --tmp-outdir-prefix={0}/output/tmp-outdir/ --tmpdir-prefix={0}/output/tmp-outdir/tmp- {1} {2}'.format('workspace',cwl_workflow_path,dataset_config_path)
     if (detached):
         log_filename = workflow + '-' + dataset + '-' + datetime.datetime.now().strftime("%Y%m%d-%H%M%S") + '.txt'
-        cwl_command = cwl_command + ' > workspace/workflow-history/' + log_filename + ' &'
+        cwl_command = cwl_command + ' > workspace/logs/' + log_filename + ' &'
     click.echo()
     click.echo(click.style('[d2s] ', bold=True) 
         + 'Running CWL worklow...')
@@ -416,7 +476,7 @@ def dataset():
         default='http://creativecommons.org/licenses/by-nc/4.0/legalcode')
 
     dataset_folder_path = 'datasets/' + dataset_id
-    os.system('cp -r d2s-cwl-workflows/support/template/dataset ' + dataset_folder_path)
+    shutil.copytree('d2s-cwl-workflows/support/template/dataset', dataset_folder_path)
     
     for dname, dirs, files in os.walk(dataset_folder_path):
         for fname in files:
@@ -433,8 +493,9 @@ def dataset():
         + click.style(dataset_id + ' dataset', bold=True) 
         + ' has been generated')
     click.echo(click.style('[d2s]', bold=True) + ' Start edit them in ' + click.style('datasets/' + dataset_id, bold=True))
-    if click.confirm(click.style('[?]', bold=True) + ' Do you want to open the ' 
-        + click.style('download', bold=True) + ' file to edit it?'):
-        os.system('nano ' + dataset_folder_path + '/download/download.sh')
+    # Will not work on all platforms:
+    # if click.confirm(click.style('[?]', bold=True) + ' Do you want to open the ' 
+    #     + click.style('download', bold=True) + ' file to edit it?'):
+    #     os.system('nano ' + dataset_folder_path + '/download/download.sh')
 
     
\ No newline at end of file
